--[[

Copyright 2008-∞ Daniel Ceregatti (daniel@ceregatti.org)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.

Special thanks go out to all the people on IRC who helped me:

Jim "jnwhiteh" Whitehead (author of TomTom)
ZorbaTHut (author of QuestHelper)
Cameron "ckknight" Knight (author of many addons)

]]--

-- Locale variable
local L = LibStub("AceLocale-3.0"):GetLocale("FriendlyArrow", false)

-- Keybindings
BINDING_HEADER_FRIENDLYARROW = L["Friendly Arrow"]
BINDING_NAME_FRIENDLYARROW_TOGGLETARGET = L["Toggle Target"]
BINDING_NAME_FRIENDLYARROW_TOGGLEFOCUS = L["Toggle Focus"]
BINDING_NAME_FRIENDLYARROW_TOGGLEMOUSEOVER = L["Toggle Mouseover"]

-- Local global variables
local gCurrentlyTracking = false
local gPlayerName = ''
local gCheckRange = 0
local gFromSlash = 0
local gTargetFrom = ''
local gRaidIndex = 0
local gGUID = ''
local gUnit = ''
local gTargetIndex = ''

local DummyOnUpdateElapsed = 0
local DummyUpdateFrequency = 1

-- Local math function aliases
local math_atan2 = math.atan2
local math_sqrt = math.sqrt

-- Local function aliases
local
  GetPlayerMapPosition, GetPlayerFacing, GetNumRaidMembers, GetNumPartyMembers, GetMouseFocus =
  GetPlayerMapPosition, GetPlayerFacing, GetNumRaidMembers, GetNumPartyMembers, GetMouseFocus

local Astrolabe = DongleStub("Astrolabe-0.4")

FriendlyArrow = LibStub("AceAddon-3.0"):NewAddon("FriendlyArrow", "AceConsole-3.0", "AceEvent-3.0")

-- Local functions
local function pp(text, debug)
  if FriendlyArrow.db.profile.ShowChatText then
    if not debug then
      FriendlyArrow:Print (text)
    --[===[@debug@
    else
      if FriendlyArrow.db.profile.ShowDebugText then
        FriendlyArrow:Print ("|cffff0000(DEBUG)|r " .. text)
      end
    --@end-debug@]===]
    end
  end
end

local function ResetAll ()
  gCurrentlyTracking = false
  gPlayerName = ''
  gCheckRange = 0
  gFromSlash = 0
  gTargetFrom = ''
  gRaidIndex = 0
  gGUID = ''
  gUnit = ''
  gTargetIndex = ''
end

local function RemoveTracking (text)
  TomTom:ReleaseCrazyArrow()
  pp (format (L["Stopped tracking %s member %s"], gTargetFrom, gPlayerName))
  pp (text, true)
  ResetAll ()
end

-- With help from Astrolabe
local function GetBearing (px, py, tx, ty)
  -- "py - ty" to account for the fact that the	0,0 origin is at the TOP left in WoW
  return -GetPlayerFacing() - math_atan2 (tx - px, py - ty)
end

local function GetRaidIndex (GUID)
  for i = 1, GetNumRaidMembers() do
    if GUID == UnitGUID("raid"..i) then
      return i
    end
  end
  return 0
end

local function GetPartyIndex (GUID)
  for i = 1, GetNumPartyMembers() do
    if GUID == UnitGUID("party"..i) then
      return i
    end
  end
  return 0
end

local function GetUnitIndex (GUID)
  local RaidIndex = GetRaidIndex(GUID)
  local TargetFrom = ''
  if RaidIndex > 0 then
    TargetFrom = 'raid'
  else
    RaidIndex = GetPartyIndex (GUID)
    if RaidIndex > 0 then
      TargetFrom = 'party'
    end
  end
  --[===[@debug@
  pp (format ("GetUnitIndex() returns: %d, %s", RaidIndex, TargetFrom), true)
  --@end-debug@]===] 
  return RaidIndex, TargetFrom
end

function FriendlyArrow:OnInitialize()
  self.db = LibStub("AceDB-3.0"):New("FriendlyArrowDB", FriendlyArrowDefaultConfigDB, "Default")

  LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("FriendlyArrow", FriendlyArrowConfig)

  self.optionsFrames = {}
  local ACD3 = LibStub("AceConfigDialog-3.0")
  self.optionsFrames.FriendlyArrow = LibStub("LibAboutPanel").new (nil, "FriendlyArrow")

  -- Thanks, Omen!
  self.optionsFrames.General = ACD3:AddToBlizOptions("FriendlyArrow", L["General"], "FriendlyArrow", "General")

  self:RegisterChatCommand("friendlyarrow", function()
    if arg == 'debug' then
      FriendlyArrow.db.profile.ShowDebugText = not FriendlyArrow.db.profile.ShowDebugText
    else
      InterfaceOptionsFrame_OpenToCategory(self.optionsFrames.General)
      InterfaceOptionsFrame_OpenToCategory(self.optionsFrames.FriendlyArrow)
    end
  end)
  self:RegisterChatCommand("fa", function(arg)
    if arg == 'debug' then
      FriendlyArrow.db.profile.ShowDebugText = not FriendlyArrow.db.profile.ShowDebugText
    else
      InterfaceOptionsFrame_OpenToCategory(self.optionsFrames.General)
      InterfaceOptionsFrame_OpenToCategory(self.optionsFrames.FriendlyArrow)
    end
  end)
  self:RegisterChatCommand("faset", function(arg)
    local result = arg:gmatch ("%S+")
    local unit = result()
    local range = result()
    FriendlyArrow:SetTracking (unit, range, 1, "")
  end)
  self.DummyFrame = CreateFrame("Frame", "FriendlyArrowDummyFrame", UIParent)
  self.DummyFrame:SetPoint("CENTER")
  self.DummyFrame:SetWidth(32)
  self.DummyFrame:SetHeight(32)
  self.DummyFrame:SetScript("OnEvent",function(frame,event,...) self[event](self,...) end)
  self.DummyFrame:RegisterEvent("RAID_ROSTER_UPDATE")
end

function FriendlyArrow:RAID_ROSTER_UPDATE()
  if gCurrentlyTracking then 
    --[===[@debug@
    pp (format ("RAID_ROSTER_UPDATE fired. gGUID: %s, gRaidIndex: %d", gGUID, gRaidIndex), true)
    --@end-debug@]===]
    FriendlyArrow:SetTracking (gUnit, gCheckRange, gFromSlash, gGUID, true)
  end
end

function FriendlyArrow:DummyOnUpdate(arg)
  DummyOnUpdateElapsed = DummyOnUpdateElapsed + arg
  if (DummyOnUpdateElapsed > DummyUpdateFrequency) then
    if UnitInRange (gTargetIndex) then
      if gFromSlash == 0 and not FriendlyArrow.db.profile.StickyKeyToggle then
        RemoveTracking ("Removed from OnUpdate with gFromSlash == 0")
      end
      if gFromSlash == 1 then
        RemoveTracking ("Removed from OnUpdate with gFromSlash == 1")
      end
    end
    DummyOnUpdateElapsed = 0
  end
end

function FriendlyArrow:SetTracking(Unit, CheckRange, FromSlash, pGUID, FromRosterUpdate)
  local GUID = pGUID or ''
  if GUID == '' then
    --[===[@debug@
    pp (format ("pGUID was empty, trying UnitGUID(Unit) for %s", Unit), true)
    --@end-debug@]===]
    GUID = UnitGUID(Unit) or ''
  end
  if GUID == '' and Unit == 'mouseover' then
    -- UnitGUID doesn't return a value for mouseover targets that are not in range!
    --[===[@debug@
    pp (format ("UnitGUID(Unit) for %s was empty, trying UnitGUID (SecureButton_GetModifiedUnit (GetMouseFocus()))", Unit), true)
    --@end-debug@]===]
    Unit = SecureButton_GetModifiedUnit (GetMouseFocus()) or 'mouseover'
    if Unit then
      GUID = UnitGUID (Unit) or ''
    end
  end
  -- Still no GUID, and we're not going to get one.
  if GUID == '' then
    -- Explicit toggles with old tracking and no new tracking removes tracking
    if FromSlash == 0 and gCurrentlyTracking then
      RemoveTracking ("Explicit toggle off")
      return
    --[[ too spammy
    else
      pp (L["Invalid or no unit"])
    ]]--
    end
  end
  -- Check to see if we've toggled ourselves
  if GUID == UnitGUID("player") then
    --[===[@debug@
    pp ("GUID is the player", true)
    --@end-debug@]===]
    return
  end
  -- At this point we have a valid GUID that is not the player
  local RaidIndex, TargetFrom = GetUnitIndex (GUID)
  Unit = TargetFrom .. RaidIndex
  -- This is an explicit toggle if it's not from a slash command or a roster update
  if FromSlash == 0 and not FromRosterUpdate then
    -- The new unit exists and is the same unit as the one we're currently tracking. Stop tracking if this was not triggered
    -- from roster update event handler and return
    if RaidIndex > 0 and RaidIndex == gRaidIndex and TargetFrom == gTargetFrom and not FromRosterUpdate then
      RemoveTracking ("Explicit toggle of the same unit as before")
      return
    end
  end
  local InRange = UnitInRange (Unit) or 0
  --[===[@debug@
  if InRange == 1 then
    pp (format ("Unit %s is in range", Unit), true)
  else
    pp (format ("Unit %s is NOT in range", Unit), true)
  end
  -- Range check toggling.
  pp (format ("Unit: %s, GUID: %s, gGUID: %s, CheckRange: %d, InRange: %d, FromSlash: %d, RaidIndex: %d, TargetFrom: %s",
    Unit, GUID, gGUID, CheckRange, InRange, FromSlash, RaidIndex, TargetFrom), true)
  --@end-debug@]===]
  if
    tonumber (CheckRange) == 1 and InRange == 1 and gFromSlash == 1
    and FriendlyArrow.db.profile.ToggleOnRange and gCurrentlyTracking
    and not FromRosterUpdate
  then
    RemoveTracking ("In range while range checking on")
    return
  end
  -- We're tracking the same person and it's from a slash command or a roster update, so there's no reason to stop tracking
  if GUID == gGUID and not FromRosterUpdate then
    --[===[@debug@
    pp (format ("Same person!"), true)
    --@end-debug@]===]
    return
  end
  -- FromSlash and gFromSlahs will never be the same in a roster update
  if FriendlyArrow.db.profile.StickyKeyToggle and FromSlash == 1 and gFromSlash == 0 and gCurrentlyTracking then
    --[===[@debug@
    pp (format ("Sticky tracking enabled while tracking and new unit is from a slash command"), true)
    --@end-debug@]===]
    return
  end
  -- Execution needs to contine past here if it's a roster update
  if tonumber (CheckRange) == 1 and InRange == 1 and FriendlyArrow.db.profile.ToggleOnRange and not FromRosterUpdate then
    --[===[@debug@
    pp (format ("The unit is in range while range tracking is set. Not tracking."), true)
    --@end-debug@]===]
    return
  end
  -- There is a valid unit to track and we should track it at this point
  gUnit = Unit
  gRaidIndex = RaidIndex
  gTargetFrom = TargetFrom
  gGUID = GUID
  gCheckRange = CheckRange
  gFromSlash = FromSlash
  gTargetIndex = TargetFrom .. RaidIndex
  -- Get the target's name
  gPlayerName, _ = UnitName (Unit)
  if gPlayerName == nil then
    -- Somehow this fails, just return
    gPlayerName = L["Unknown Player?"]
  end
  -- The tracked target will be in the same zone as us
  local c, z = GetCurrentMapContinent (), GetCurrentMapZone ()
  -- GetPlayerMapPosition() will return nil on untrackable units (not the same zone/bg/instance)
  if select (1, GetPlayerMapPosition(gTargetIndex)) ~= 0 then
    local DistanceCalculator
    local DistanceMetric = L["yards"]
    if c == -1 and z == 0 then
      local height = FriendlyArrowGetInstanceMapWidth ()
      local width = height * 1.5
      --[===[@debug@
      pp (format ("Using local distance calculator. Width: %d, Height: %d", width, height), true)
      --@end-debug@]===]
      DistanceCalculator = function () return math_sqrt ((((tx - px) * width) ^ 2) + (((ty - py) * height) ^ 2)) end
      if width == 100 then
        DistanceMetric = L["coordinates"]
      end
    else
      --[===[@debug@
      pp ("Using Astrolabe distance calculator", true)
      --@end-debug@]===]
      DistanceCalculator = function () return select (1, Astrolabe:ComputeDistance (c, z, px, py, c, z, tx, ty)) end
    end
    if not FromRosterUpdate then
      pp (format (L["Tracking %s member %s"], gTargetFrom, gPlayerName))
    end
    if FriendlyArrow.db.profile.ToggleOnRange then
      FriendlyArrow.DummyFrame:SetScript("OnUpdate", FriendlyArrow.DummyOnUpdate)
    end
    gCurrentlyTracking = true
    TomTom:SetCrazyArrowColor(0, 1, 0)
    TomTom:HijackCrazyArrow(function(self, elapsed)
      tx, ty = GetPlayerMapPosition(gTargetIndex)
      px, py = GetPlayerMapPosition("player")
      TomTom:SetCrazyArrowDirection(GetBearing (px, py, tx, ty))
      TomTom:SetCrazyArrowTitle(gPlayerName, format ("%d " .. DistanceMetric, DistanceCalculator()), "")
      if tx == 0 and ty == 0 then
        RemoveTracking("GetPlayerMapPosition() returned 0,0")
      end
    end)
  else
    --[[ also spammy
    local Reason
    if GetNumRaidMembers() == 0 or GetNumPartyMembers() == 0 then
      Reason = L["You are not in raid or party"]
    elseif RaidIndex == 0 then
      Reason = L["Unit not in your raid or party"]
    else
      local Location = L["zone"]
      local InInstance, InstanceType = IsInInstance()
      if InInstance then
        if InstanceType == 'pvp' then
          Location = L["battleground"]
        elseif InstanceType == 'arena' then
          Location = L["arena"]
        elseif InstanceType == 'party' then
          Location = L["instance"]
        elseif InstanceType == 'raid' then
          Location = L["raid instance"]
        end
      end
      Reason = format (L["Unit is not in the same %s"], Location)
    end
    pp (format (L["%s cannot be tracked: %s"], gPlayerName, Reason))
    ]]--
    ResetAll()
  end
end